from dataclasses import dataclass
from datetime import datetime
from typing import List

from .querytypes.const import (
HOURLY,
QUARTER_HOURLY
)

from .querytypes.querytypes import(
LocaleQuery
)

@dataclass(frozen=True)
class Locale_Type:
    observed_peak:str
    charged_peak:str
    query_model: LocaleQuery
    free_charge_pattern:List = None
    peak_cycle:str = HOURLY
    converted:bool = False #transition key to remove sql-dependency

    def free_charge(self, mockdt:datetime = datetime.min) -> bool:
        if self.free_charge_pattern is None or len(self.free_charge_pattern) == 0:
            return False
        now = datetime.now() if mockdt is datetime.min else mockdt
        for p in self.free_charge_pattern:
            if now.month in p["M"]:
                if now.weekday() in p["D"]:
                    if now.hour in p["H"]:
                        return True
        return False

    def is_quarterly(self) -> bool:
        return self.peak_cycle == QUARTER_HOURLY





import time

class PowerReading:
    def __init__(self, price, power, td) -> None:
        self._price = price
        self._power = power
        self._time_delta = td
    
    @property
    def reading_integral(self):
        return self.power/1000 * (self.time_delta/3600)

    @property
    def reading_cost(self):
        return self.reading_integral * self.price

    @property
    def price(self):
        return self._price

    @property
    def power(self):
        return self._power

    @property
    def time_delta(self):
        return self._time_delta


class SessionPrice:
    def __init__(self) -> None:
        self._total_price = 0
        self._price = 0
        self._current_power = 0
        self._total_power = 0
        self._current_time = 0
        self._time_delta = 0
        self._readings = []
    
    @property
    def total_power(self):
        return round(self._total_power,5)

    def terminate(self, mock_time:float=None):
        print("called terminate")
        self.update_power_reading(0, mock_time)
        self.get_status()
        
    def get_status(self) -> dict:
        self._total_power = 0
        self._total_price = 0
        for i in self._readings:
            self._total_power += i.reading_integral
            self._total_price += i.reading_cost
        return {
            "energy": {
                "value": self._total_power,
                "unit": "kWh"
            },
            "price": self._total_price
        }
            
    def update_power_reading(self, power:float, mock_time:float=None):
        self._set_delta(mock_time)
        p = PowerReading(self._price, self._current_power, self._time_delta)
        print(p.power, p.price, p.reading_cost, p.reading_integral)
        self._readings.append(p)
        self._current_power = power
        
    def update_price(self, price:float, mock_time:float=None):
        if self._current_power > 0:
            self.update_power_reading(
                power=self._current_power, 
                mock_time=mock_time
                )
        self._price = price

    def _set_delta(self, mock_time:float=None) -> None:
        now = time.time() if mock_time is None else mock_time
        self._time_delta = (now - self._current_time)
        self._current_time = now

    @property
    def readings(self) -> list:
        return self._readings

    @property
    def total_price(self) -> float:
        return self._total_price

    @total_price.setter
    def total_price(self, val):
        self.total_price = val




import time
from datetime import datetime
from .Models import CHARGERSTATES

class ChargeControllerBase:  
    def __init__(
        self,
        charger_state_translation:dict[CHARGERSTATES,list[str]],
        non_hours:list[int] = [],
        timeout:int = 180
    ):
        self._charger_state_translation = self._check_charger_states(charger_state_translation)
        self._done_timeout = timeout
        self._latest_charger_start = time.time()
        self._non_hours = non_hours
    
    @property
    def done_timeout(self):
        return self._done_timeout

    def below_start_threshold(
            self,
            predicted_energy: float,
            current_peak: float,
            threshold_start: float
    ) -> bool:
        """
        Check if you are below the starting threshold of charger
        :param predicted_energy: predicted energy for this hour in kWh
        :param current_peak: currently set energy peak, in kWh
        :param threshold_start: the allowed peak percentage to start. 0-1
        :return: returns bool
        """
        return (predicted_energy * 1000) < ((current_peak * 1000) * threshold_start)

    def above_stop_threshold(
            self,
            predicted_energy: float,
            current_peak: float,
            threshold_stop: float
    ) -> bool:
        """
        Check if you are above the stopping threshold of charger
        :param predicted_energy: predicted energy for this hour in kWh
        :param current_peak: currently set energy peak, in kWh
        :param threshold_stop: the limit peak percentage to stop. 0-1
        :return: returns bool
        """
        return (predicted_energy * 1000) > ((current_peak * 1000) * threshold_stop)

    # @property
    # def latest_charger_start(self):
    #     return self._latest_charger_start

    # @latest_charger_start.setter
    # def latest_charger_start(self, val=None):
    #     self._latest_charger_start = time.time()
    
    @property
    def _is_timeout(self) -> bool:
        return time.time() - self.latest_charger_start > self._done_timeout

    # def get_status(
    #     self,
    #     charger_state:str,
    #     charger_enabled:bool,
    #     charger_done:bool,
    #     car_power_sensor:float,
    #     total_energy_this_hour:float,
    #     current_hour:int|None
    #     ) -> CHARGERSTATES:
    #     _update_timer = False
    #     self._charger_state = charger_state.lower()
    #     self._current_hour = current_hour if current_hour is not None else datetime.now().hour
    #     ret = CHARGERSTATES.Error

    #     if self._charger_state in self._charger_state_translation[CHARGERSTATES.Idle]:
    #         _update_timer = True
    #         ret = CHARGERSTATES.Idle

    #     elif self._charger_state in self._charger_state_translation[CHARGERSTATES.Connected] and charger_enabled is False:
    #         _update_timer = True
    #         ret = CHARGERSTATES.Connected

    #     elif self._charger_state not in self._charger_state_translation[CHARGERSTATES.Idle] and charger_done is True:
    #         ret = CHARGERSTATES.Done

    #     elif self._current_hour in self._non_hours:
    #         _update_timer = True
    #         ret = CHARGERSTATES.Stop

    #     elif self._charger_state in self._charger_state_translation[CHARGERSTATES.Connected]:
    #         if car_power_sensor < 1 and self._is_timeout:
    #             ret = CHARGERSTATES.Done
    #         else:
    #             if self.below_start_threshold and total_energy_this_hour > 0:
    #                 ret = CHARGERSTATES.Start
    #             else:
    #                 _update_timer = True
    #                 ret = CHARGERSTATES.Stop

    #     elif self._charger_state in self._charger_state_translation[CHARGERSTATES.Charging]:
    #         _update_timer = True
    #         if self.above_stop_threshold and total_energy_this_hour > 0:
    #             ret = CHARGERSTATES.Stop
    #         else:
    #             ret = CHARGERSTATES.Start

    #     if _update_timer is True:
    #         self.latest_charger_start = 1
    #     return ret

    def _check_charger_states(self, input:dict[CHARGERSTATES,list[str]]) -> dict[CHARGERSTATES,list[str]]:
        if len(input) == 0:
            raise AssertionError
        for i in input:
            if len(input[i]) == 0:
                raise AssertionError
        return input


from ..locale.querytypes.const import (
    QUERYTYPE_BASICMAX
)

from ..locale.locale_model import Locale_Type
from dataclasses import dataclass

@dataclass(frozen=True)
class Default(Locale_Type):
    observed_peak = QUERYTYPE_BASICMAX
    charged_peak = QUERYTYPE_BASICMAX
    converted = True



from datetime import date, datetime, time
import logging

from .const import (
    QUERYTYPE_AVERAGEOFTHREEHOURS,
    QUERYTYPE_AVERAGEOFTHREEDAYS,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19,
    QUERYTYPE_BASICMAX,
    QUERYTYPE_AVERAGEOFFIVEDAYS,
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR,
    QUERYTYPE_HIGHLOAD,
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22,
    QUERYTYPE_SOLLENTUNA
    )
from .queryservice import QueryService
from .models.peaks_model import PeaksModel
from .models.enums import SumTypes, TimePeriods
from .models.sumcounter import SumCounter
from .models.queryproperties import QueryProperties

from .querysets import QUERYSETS

_LOGGER = logging.getLogger(__name__)


class LocaleQuery:
    def __init__(
        self, 
        sumtype: SumTypes, 
        timecalc: TimePeriods, 
        cycle: TimePeriods, 
        sumcounter: SumCounter = None,
        queryservice: QueryService = QueryService()
        ) -> None:    
        self._peaks:PeaksModel = PeaksModel({})
        self._props = QueryProperties(
            sumtype, 
            timecalc, 
            cycle,
            queryservice
            )
        self._sumcounter:SumCounter= sumcounter
        self._observed_peak_value:float = 0 
        self._charged_peak_value:float = 0

    def reset(self) -> None:
        self._peaks.reset()
        self._observed_peak_value = 0
        self._charged_peak_value = 0

    @property
    def peaks_export(self) -> dict:
        ppdict = {}
        for pp in self._peaks.p:
            ppkey = str(str(pp[0]) + "h" + str(pp[1]))
            ppdict[ppkey] = self._peaks.p[pp]
        return {
            "m": self._peaks.m,
            "p": ppdict
        }

    @property
    def peaks(self) -> PeaksModel:
        if self._peaks.is_dirty:
            self._sanitize_values()
        return self._peaks

    @property
    def sumcounter(self) -> SumCounter:
        if self._sumcounter is not None:
            return self._sumcounter
        return SumCounter()

    @property
    def charged_peak(self) -> float: 
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = self._charged_peak_value
        return round(ret,2)

    @charged_peak.setter
    def charged_peak(self, val):
        self._charged_peak_value = val

    @property
    def observed_peak(self) -> float: 
        if self._peaks.is_dirty:
            self._sanitize_values()
        ret = self.charged_peak if self._props.sumtype is SumTypes.Max else self._observed_peak_value
        return round(ret, 2)

    @observed_peak.setter
    def observed_peak(self, val):
        self._observed_peak_value = val

    def try_update(self, newval, timestamp:datetime = datetime.now()):
        if self._props.queryservice.should_register_peak(dt=timestamp) is False:
            return
        if self.peaks.is_dirty:
            self._sanitize_values()
        _dt = (timestamp.day, timestamp.hour)
        if len(self.peaks.p) == 0:
            """first addition for this month"""
            self._peaks.p[_dt] = newval
            self._peaks.m = timestamp.month
        elif timestamp.month != self._peaks.m:
            """new month, reset"""
            self.reset_values(newval, timestamp)
        else:
            self._set_update_for_groupby(newval, _dt)
        if len(self.peaks.p) > self.sumcounter.counter:
                self.peaks.p.pop(min(self.peaks.p, key=self._peaks.p.get))
        self._update_peaks()

    def _set_update_for_groupby(self, newval, _dt):
        if self.sumcounter.groupby in [TimePeriods.Daily, TimePeriods.UnSet]:
            _datekey = [k for k,v in self.peaks.p.items() if _dt[0] in k]
            if len(_datekey) > 0:
                if newval > self.peaks.p[_datekey[0]]:
                        self.peaks.p.pop(_datekey[0])
                        self.peaks.p[_dt] = newval
            else:
                self.peaks.p[_dt] = newval
        elif self.sumcounter.groupby == TimePeriods.Hourly:
            if _dt in self._peaks.p.keys():
                if newval > self.peaks.p[_dt]:
                        self.peaks.p[_dt] = newval
            else:
                self.peaks.p[_dt] = newval

    def _update_peaks(self):
        if self._props.sumtype is SumTypes.Max:
            self.charged_peak = max(self._peaks.p.values())
        elif self._props.sumtype is SumTypes.Avg:
            self.observed_peak = min(self._peaks.p.values())
            self.charged_peak = sum(self._peaks.p.values()) / len(self._peaks.p)

    def reset_values(self, newval, dt = datetime.now()):
        self._peaks.p.clear()
        self.try_update(newval, dt)

    def _sanitize_values(self):
        def countX(lst, x):
            count = 0
            for ele in lst:
                if ele[0] == x:
                    count = count + 1
            return count
        if self.sumcounter.groupby == TimePeriods.Daily:
            duplicates = {}
            for k in self._peaks.p.keys():
                if countX(self._peaks.p.keys(), k[0]) > 1:
                    duplicates[k] = self._peaks.p[k]
            if len(duplicates) > 0:
                minkey = min(duplicates, key=duplicates.get)
                self._peaks.p.pop(minkey)
                    
        while len(self._peaks.p) > self.sumcounter.counter:
            self._peaks.p.pop(min(self._peaks.p, key=self._peaks.p.get))
        self._peaks.is_dirty = False
        self._update_peaks()

QUERYTYPES = {
    QUERYTYPE_AVERAGEOFTHREEHOURS: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Hourly)),
    QUERYTYPE_AVERAGEOFTHREEDAYS: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Daily)),
    QUERYTYPE_BASICMAX: LocaleQuery(sumtype=SumTypes.Max, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly),
    QUERYTYPE_AVERAGEOFFIVEDAYS: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=5, groupby=TimePeriods.Daily)),
    QUERYTYPE_SOLLENTUNA: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Hourly), queryservice=QueryService(QUERYSETS[QUERYTYPE_SOLLENTUNA])),
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22: LocaleQuery(sumtype=SumTypes.Max, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, queryservice=QueryService(QUERYSETS[QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22])),
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19: LocaleQuery(sumtype=SumTypes.Avg, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, sumcounter=SumCounter(counter=3, groupby=TimePeriods.Hourly), queryservice=QueryService(QUERYSETS[QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19])),
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR: LocaleQuery(sumtype=SumTypes.Max, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, queryservice=QueryService(QUERYSETS[QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR])),
    QUERYTYPE_HIGHLOAD: LocaleQuery(sumtype=SumTypes.Max, timecalc=TimePeriods.Hourly, cycle=TimePeriods.Monthly, queryservice=QueryService(QUERYSETS[QUERYTYPE_HIGHLOAD]))
}



# from .Threshold import ThresholdBase as Threshold
# from .Chargecontroller import ChargeControllerBase as ChargeController
# from .Prediction import PredictionBase as Prediction
# from .hoursselection import Hoursselectionbase as Hoursselection
# from .session import SessionPrice as Session
from .locale.querytypes import querytypes
from .locale import (Locale, locale_model)
from .country import (belgium, sweden, norway, default)





from dataclasses import dataclass
from datetime import datetime

@dataclass
class PeaksModel:
    p: dict
    m:int = 0
    is_dirty:bool = False

    def set_init_dict(self, dict_data, dt = datetime.now()):
        if dt.month == self.m:
            ppdict = {}
            for pp in dict_data["p"]:
                tkeys = pp.split("h")
                ppkey = (int(tkeys[0]), int(tkeys[1]))
                ppdict[ppkey] = dict_data["p"][pp]
            if len(self.p) > 0:
                ppdict = self.p | ppdict
            self.p = ppdict
            self.m = dict_data["m"]
            self.is_dirty = True

    def reset(self) -> None:
        self.m = 0
        self.is_dirty = False
        self.p = {}


import setuptools

setuptools.setup(
    name="peaqevcore",
    version="1.1.2",
    author="Magnus Eldén",
    description="Core types for peaqev car charging",
    url="https://github.com/elden1337/peaqev-core",
    license="MIT",
    packages=[
        "peaqevcore", 
        "peaqevcore.locale", 
        "peaqevcore.country",
        "peaqevcore.hourselection", 
        "peaqevcore.locale.querytypes", 
        "peaqevcore.locale.querytypes.models"
        ],
    test_requires=['pytest']
)   
